@@ -4,7 +4,7 @@ from PyQt5.QtGui import QPainter,QPainterPath
 from PyQt5 import QtGui
 from PyQt5.QtCore import Qt
 from scipy.interpolate import interp1d
-
+import time
 
 import numpy as np
 # from transaction_editor import TransactionEditor
@@ -86,10 +86,24 @@ class Box:
 
 
 class Connector:
-    def __init__(self,a,b,name="Connector"):
+    def __init__(self,a,b,name="Connector",items=None):
+        """
+        creates a new connector
+
+        :param a: a box (a) sender
+        :param b: a box (b) receiver
+        :param name: the label of the conncetor
+        :param items: items affected by this connector (list of str)
+        """
+
         self.a = a
         self.b = b
 
+        if items is None:
+            self.items = []
+        else:
+            self.items =items
+
         #self.a.n_connections += 1
         #if self.a != self.b:
         #    self.b.n_connections += 1
@@ -102,12 +116,13 @@ class Connector:
         v = np.random.uniform(-1,1)
 
         if a != b:
-
             self.c = Box(self.a.x + 0.5*dx + u, self.a.y + 0.5*dy + v,name,ishelper=True,parent=self)
             # self.d = Box(self.a.x + 0.75*dx, self.a.y + 0.75*dy,None,ishelper=True,parent=self)
         else:
             self.c = Box(self.a.x + 5*u, self.a.y + 5*u ,name,ishelper=True,parent=self)
 
+        self.a.ishelper = False
+        self.b.ishelper = False
 
 class MyDrawWidget(QtWidgets.QWidget):
 
@@ -134,6 +149,7 @@ class MyDrawWidget(QtWidgets.QWidget):
         self.connected2 = None
         self._highlighted = None
 
+        self.last_click = time.time()
 
         self.parent().zoomOutButton.pressed.connect(self.zoom_out)
         self.parent().zoomInButton.pressed.connect(self.zoom_in)
@@ -161,12 +177,23 @@ class MyDrawWidget(QtWidgets.QWidget):
         self.mousex = 0
         self.mousey = 0
 
+        self.interpol_method = "cubic"
         self.mode = "select"
 
         self.zoom_factor = 1.0
 
         self.show()
-
+    
+    
+    def swap_interpol_style(self):
+        if self.interpol_method == "cubic":
+            self.interpol_method = "quadratic"
+        elif self.interpol_method == "quadratic":
+            self.interpol_method = "slinear"
+        elif self.interpol_method == "slinear":
+            self.interpol_method = "cubic"
+        self.parent().update()
+    
     def wheelEvent(self,event):
 
         numDegrees = event.angleDelta().y()
@@ -189,14 +216,16 @@ class MyDrawWidget(QtWidgets.QWidget):
     def zoom_out(self):
 
         self.zoom_factor -= 0.1
-        if self.zoom_factor < 1.0:
-            self.zoom_factor = 1.0
+        if self.zoom_factor < .8:
+            self.zoom_factor = .8
 
         self.update()
 
     def resize_frame(self):
         self.W = max(500,int(self.parent().frameGeometry().width()-970)) # 850
         self.H = max(850,int(850+self.parent().frameGeometry().height()-986)) # 750
+
+
         self.maxx = self.W- 50 # 800
         self.maxy = self.H- 50 # 700
         #self.setGeometry(890,170,self.W,self.H)
@@ -220,8 +249,10 @@ class MyDrawWidget(QtWidgets.QWidget):
                 if val.n_connections == 0:
                     self.parent().graphDeleteButton.setEnabled(True)
                 else:
+                    print("Delete disabled because has still %i connections" % val.n_connections)
                     self.parent().graphDeleteButton.setEnabled(False)
             else:
+                print("Cannot delete (is helper)")
                 self.parent().graphEditButton.setEnabled(False)
                 self.parent().graphDeleteButton.setEnabled(False)
 
@@ -264,6 +295,9 @@ class MyDrawWidget(QtWidgets.QWidget):
             self.parent().notify("Box has still %i connection(s)! Cannot remove"%box.n_connections, title="Error")
 
     def check_exist(self,name):
+        if name == "":
+            return False
+
         for box in self.boxes:
             # TODO more efficient search for large number of boxes?
             if box.name == name:
@@ -303,13 +337,13 @@ class MyDrawWidget(QtWidgets.QWidget):
 
         self.update()
 
-    def add_connection(self,box1, box2, name):
+    def add_connection(self,box1, box2, name,items):
 
         #for box in self.boxes:
         #    if box.ishelper and box.name == name:
         #        return
 
-        new_conn = Connector(box1,box2,name)
+        new_conn = Connector(box1,box2,name,items)
 
         if name not in self.old_positions:
             self.old_positions[name] = new_conn.c.x, new_conn.c.y
@@ -429,11 +463,13 @@ class MyDrawWidget(QtWidgets.QWidget):
 
 
 
-    def curved_connector(self,qp,x0,y0,xf,yf,xi,yi):
+    def curved_connector(self,qp,x0,y0,xf,yf,xi,yi,dashed=False):
         """
         draws a curved connector
         """
 
+        #x0 += 12.5
+        #y0 += 12.5
         # construct helper box_positions
         xi2 = 0.5*(xi-x0) + x0
         xi3 = 0.5*(xf-xi) + xi
@@ -442,7 +478,7 @@ class MyDrawWidget(QtWidgets.QWidget):
 
         #qp.drawRect(xi2-5,yi2-5,10,10)
         #qp.drawRect(xi3-5,yi3-5,10,10)
-        W = 12.5
+        W = 25 # 25 # 12.5
 
         xi4 = xi
         yi4 = yi
@@ -547,13 +583,19 @@ class MyDrawWidget(QtWidgets.QWidget):
 
         # https://stackoverflow.com/questions/52014197/how-to-interpolate-a-2d-curve-in-python
         # define some points
+        end_is_start = False
         if abs(x0-xf) < 0.1 and abs(y0-yf) < 0.1:
+            end_is_start = True
             points2 = np.array([[x0,xi6,xi62,xi2,xi],
                                 [y0,yi6,yi62,yi2,yi]]).T
         else:
             points2 = np.array([[x0,xi6,xi62,xi2,xi,xi3,xi35,xi5,xf],
                           [y0,yi6,yi62,yi2,yi,yi3,yi35,yi5,yf]]).T
 
+            points2 = np.array([[x0,xi2,xi,xi3,xf],
+                          [y0,yi2,yi,yi3,yf]]).T
+
+
         points = []
         points.append(points2[0]) # make sure it always starts at x0,y0 and ends at xf, yf
 
@@ -572,20 +614,77 @@ class MyDrawWidget(QtWidgets.QWidget):
 
             # Interpolation for different methods:
             #interpolations_methods = ['slinear', 'quadratic', 'cubic']
-            method  = "quadratic"
-            alpha = np.linspace(0, 1, 75)
-
+            if end_is_start:
+                method = "quadratic"
+                alpha = np.linspace(0, 1, 50)
+            else:
+                method  = self.interpol_method 
+                alpha = np.linspace(0, 1, 80)
 
             #for method in interpolations_methods:
             interpolator =  interp1d(distance, points, kind=method, axis=0)
             interpolated_points= interpolator(alpha)
 
+            # draw arrow at mid point
+            n = len(interpolated_points)
+
             # print("interpolated_poitns",interpolated_points)
-            for i in range(1,len(interpolated_points)):
+            # draw curved line
+            offset_text = 20
+
+            for i in range(1,int(n/2)-offset_text):
                 qp.drawLine(self.zoom_factor*(interpolated_points[i-1][0]),self.zoom_factor*(interpolated_points[i-1][1]),
                             self.zoom_factor*(interpolated_points[i][0]),self.zoom_factor*(interpolated_points[i][1]))
+
+            for i in range(int(n/2)-offset_text,len(interpolated_points)):
+                qp.drawLine(self.zoom_factor*(interpolated_points[i-1][0]),self.zoom_factor*(interpolated_points[i-1][1]),
+                            self.zoom_factor*(interpolated_points[i][0]),self.zoom_factor*(interpolated_points[i][1]))
+
+
+            mid_point = interpolated_points[int(n/2)]
+            mid_point2 = interpolated_points[int(n/2)+1]
+            mid_points3 = interpolated_points[int(n/2)-offset_text+1]
+
+            dy = mid_point[1] - mid_point2[1]
+            dx = mid_point[0] - mid_point2[0]
+
+            alfa = np.arctan2(dy,dx)
+            beta1 = alfa + np.pi/4
+            beta2 = alfa - np.pi/4
+
+            if abs(dy) > abs(dy):
+                beta3 = alfa - np.pi/3
+            else:
+                beta3 = alfa - np.pi/3 + np.pi/2
+
+            x2 = mid_point2[0]
+            y2 = mid_point2[1]
+
+            d = 10
+            x3 = np.cos(beta1)*d + x2
+            y3 = np.sin(beta1)*d + y2
+
+            x4 = np.cos(beta2)*d + x2
+            y4 = np.sin(beta2)*d + y2
+
+            xt = np.cos(beta3)*3 + mid_points3[0]
+            yt = np.sin(beta3)*3 + mid_points3[1]
+
+
+            if not end_is_start and self.parent().checkBox.isChecked():
+                # left side
+                qp.drawLine(self.zoom_factor*(x2), self.zoom_factor*(y2), self.zoom_factor*(x3), self.zoom_factor*(y3))
+
+                # right side
+                qp.drawLine(self.zoom_factor*(x2), self.zoom_factor*(y2), self.zoom_factor*(x4), self.zoom_factor*(y4))
+
+
+
+
         except Exception as e:
             print("Error in Line Interpolation:",str(e))
+            xt = 0
+            yt = 0
 
         """
         if self.darkMode:
@@ -611,6 +710,8 @@ class MyDrawWidget(QtWidgets.QWidget):
         #self.draw_bezier(qp,xi,yi,xi3,yi3,xi7,yi7) # starting point
         #self.draw_bezier(qp,xi3,yi3,xf,yf,xi5,yi5) # starting point
 
+        return xt,yt
+
     def dawPreview(self,qp):
         # draw small preview window
         # param qp: a qpainter object
@@ -629,7 +730,7 @@ class MyDrawWidget(QtWidgets.QWidget):
         move_y = -10 # -25
 
         brush = QtGui.QBrush(self.parent().theme_manager.get_color(7))
-        qp.fillRect(QtCore.QRect(move_x + self.W0-distance-w_frame,move_y + self.H0-distance-h_frame,w_frame+25,h_frame+25),brush)
+        qp.fillRect(QtCore.QRect(move_x + self.W0-distance-w_frame,move_y + self.H-distance-h_frame,w_frame+25,h_frame+25),brush)
 
         if len(self.boxes) == 0:
             return
@@ -688,11 +789,10 @@ class MyDrawWidget(QtWidgets.QWidget):
         y_vals = [y+0.5*h_frame for y in y_vals]
 
         x_vals = [move_x + self.W0-w_frame+x-2-10 for x in x_vals]
-        y_vals = [move_y + self.H0-h_frame+y-2-10 for y in y_vals]
+        y_vals = [move_y + self.H-h_frame+y-2-10 for y in y_vals]
 
         # transform all points
 
-
         for i in range(len(x_vals)-2):
             x = x_vals[i]
             y = y_vals[i]
@@ -708,7 +808,6 @@ class MyDrawWidget(QtWidgets.QWidget):
                 #else:
                 qp.drawRect(x,y,4,4)
 
-
         # skip origin ...
         x = x_vals[-1]
         y = y_vals[-1]
@@ -749,17 +848,34 @@ class MyDrawWidget(QtWidgets.QWidget):
 
         qp.drawRect(0,0,self.W-2,self.H-2)
 
-        #pen2 = QtGui.QPen(self.parent().theme_manager.get_color(2),1)
-        #qp.setPen(pen2)
-        #qp.drawRect(0+self.gox,0+self.goy,self.gox + self.W-2,self.goy + self.H-2)
+        # pen2 = QtGui.QPen(self.parent().theme_manager.get_color(2),1)
+        # qp.setPen(pen2)
+        # qp.drawRect(0+self.gox,0+self.goy,self.gox + self.W-2,self.goy + self.H-2)
 
         qp.setPen(pen)
 
+        filter_vals = set([item.text() for item in self.parent().filterCombo.selectedItems()])
+        active_boxes = []
+        active_connectors = []
+
+        for conn in self.connectors:
+            no_selection = len(filter_vals) == 0
+            is_selected = len(filter_vals) != 0 and len(list(set(conn.items) & filter_vals)) > 0
+            if no_selection or is_selected:
+                if conn not in active_connectors:
+
+                    if self.highlighted is None:
+                        active_connectors.append(conn)
+                    else:
+                        if conn.a == self.highlighted or conn.b == self.highlighted or conn.c == self.highlighted:
+                            active_connectors.append(conn)
+
+
         for conn in self.connectors:
 
             if conn.c == self.highlighted:
-                pen = QtGui.QPen(self.parent().theme_manager.get_color(4),2.0) # ,2, Qt.SolidLine)
 
+                pen = QtGui.QPen(self.parent().theme_manager.get_color(4),2.0) # ,2, Qt.SolidLine)
 
             else:
                 # pen = QtGui.QPen(QtGui.QColor(0, 0, 0, 200),1) # ,2, Qt.SolidLine)
@@ -770,22 +886,84 @@ class MyDrawWidget(QtWidgets.QWidget):
                 else:
                     pen = QtGui.QPen(QtGui.QColor(0, 0,0 , 75),0) # ,2, Qt.SolidLine)
                 """
+
                 pen = QtGui.QPen(self.parent().theme_manager.get_color(2),1)
 
+                if conn in active_connectors:
+                    pen.setStyle(Qt.SolidLine)
+                    pen.setWidthF(1.0)
+
+                else:
+                    pen.setStyle(Qt.DashLine)
+                    #pen.setDashOffset(4.0)
+                    pen.setDashPattern([1,4])
+                    pen.setWidthF(0.8)
+
+
             qp.setPen(pen)
 
             # qp.drawLine(int(conn.a.x+25), int(conn.a.y+25), int(conn.c.x+5), int(conn.c.y+5))
             # qp.drawLine(int(conn.c.x+5), int(conn.c.y+5), int(conn.b.x+25), int(conn.b.y+25))
-            self.curved_connector(qp,self.gox + int(conn.a.x+12.5), self.goy + int(conn.a.y+12.5),self.gox + int(conn.b.x+12.5), self.goy + int(conn.b.y+12.5), self.gox + int(conn.c.x+5), self.goy + int(conn.c.y+5))
-            # self.curved_connector(qp,int(conn.a.x), int(conn.a.y),int(conn.b.x), int(conn.b.y),int(conn.c.x), int(conn.c.y))
 
-            if conn.c == self.highlighted:
-                pen = QtGui.QPen(self.parent().theme_manager.get_color(2),1)
-                qp.setPen(pen)
+            # apply filter: does the connector contain any of the filter_vals?
+
+            show = True
+
+
+
+            if conn in active_connectors:
+                # draw only if filter applies
+
+                offset_box = 25
+                tx,ty = self.curved_connector(qp,self.gox + int(conn.a.x+offset_box), self.goy + int(conn.a.y+offset_box),self.gox + int(conn.b.x+offset_box), self.goy + int(conn.b.y+offset_box), self.gox + int(conn.c.x+5), self.goy + int(conn.c.y+5))
+                # self.curved_connector(qp,int(conn.a.x), int(conn.a.y),int(conn.b.x), int(conn.b.y),int(conn.c.x), int(conn.c.y))
+                if conn.a not in active_boxes:
+                    active_boxes.append(conn.a)
+                if conn.b not in active_boxes:
+                    active_boxes.append(conn.b)
+                if conn.c not in active_boxes:
+                    active_boxes.append(conn.c)
+
+
+                if conn.c == self.highlighted:
+                    pen = QtGui.QPen(self.parent().theme_manager.get_color(2),1)
+                    qp.setPen(pen)
+                else:
+                    pen = QtGui.QPen(self.parent().theme_manager.get_color(1),1)
+                    qp.setPen(pen)
+
+                if self.parent().checkBox_5.isChecked(): # draw at handles
+                    if self.parent().checkBox_2.isChecked():
+                        if self.highlighted == conn.c:
+                            brush = QtGui.QBrush(self.parent().theme_manager.get_color(4),1)
+                        else:
+                            brush = QtGui.QBrush(self.parent().theme_manager.get_color(2),1)
+
+                        a = self.zoom_factor*(self.gox +conn.c.x)
+                        b = self.zoom_factor*(self.goy + conn.c.y-7)
+                        if self.parent().checkBox_4.isChecked():
+                            qp.fillRect(QtCore.QRect(QtCore.QPoint(int(a-.1),b-10),QtCore.QPoint(int(a+5.1*len(conn.c.name)),b+10)),brush)
+
+                        qp.drawText(a,b,conn.c.name)
+                else:
+                    if self.parent().checkBox_2.isChecked():
+                        if self.highlighted == conn.c:
+                            brush = QtGui.QBrush(self.parent().theme_manager.get_color(4),1)
+                        else:
+                            brush = QtGui.QBrush(self.parent().theme_manager.get_color(2),1)
+                        if conn.a.x != conn.b.x and conn.a.y != conn.b.y:
+                            a = self.zoom_factor*(tx)
+                            b = self.zoom_factor*(ty)
+                        else:
+                            a = self.zoom_factor*(self.gox +conn.c.x)
+                            b = self.zoom_factor*(self.goy + conn.c.y-7)
+                        if self.parent().checkBox_4.isChecked():
+                            qp.fillRect(QtCore.QRect(QtCore.QPoint(int(a-.1),b-10),QtCore.QPoint(int(a+5.1*len(conn.c.name)),b+10)),brush)
+                        qp.drawText(a,b,conn.c.name.replace("_"," "))
+
+
             else:
-                pen = QtGui.QPen(self.parent().theme_manager.get_color(1),1)
-                qp.setPen(pen)
-            qp.drawText(self.zoom_factor*(self.gox + conn.c.x),self.zoom_factor*(self.goy + conn.c.y-7),conn.c.name)
+                self.curved_connector(qp,self.gox + int(conn.a.x+12.5), self.goy + int(conn.a.y+12.5),self.gox + int(conn.b.x+12.5), self.goy + int(conn.b.y+12.5), self.gox + int(conn.c.x+5), self.goy + int(conn.c.y+5),dashed=True)
 
         """
         if self.darkMode:
@@ -794,6 +972,7 @@ class MyDrawWidget(QtWidgets.QWidget):
             pen = QtGui.QPen(QtGui.QColor(0, 0,0 , 75),0) # ,2, Qt.SolidLine)
         """
         pen = QtGui.QPen(self.parent().theme_manager.get_color(2),1)
+        pen.setStyle(Qt.SolidLine)
 
         #pen = QtGui.QPen(QtGui.QColor(0, 0,0 , 200),1) # ,2, Qt.SolidLine)
 
@@ -833,6 +1012,7 @@ class MyDrawWidget(QtWidgets.QWidget):
                 """
                 br = QtGui.QBrush(self.parent().theme_manager.get_color(0),1)
 
+
             elif box == self.highlighted:
                 br = QtGui.QBrush(self.parent().theme_manager.get_color(5))
             else:
@@ -849,6 +1029,7 @@ class MyDrawWidget(QtWidgets.QWidget):
 
             if box.ishelper:
                 pen = QtGui.QPen(self.parent().theme_manager.get_color(6),1)
+                br = QtGui.QBrush(self.parent().theme_manager.get_color(1),1)
                 qp.setPen(pen)
                 # qp.drawText(x,y-7,box.name)
             else:
@@ -864,11 +1045,19 @@ class MyDrawWidget(QtWidgets.QWidget):
                 x2 = x + 6
                 y2 = y + 6
 
+
                 pen = QtGui.QPen(QtGui.QColor(0, 0, 0 , 0),0) # ,2, Qt.SolidLine)
                 qp.setPen(pen)
+                # qp.drawRect(QtCore.QRect(QtCore.QPoint(x,y),QtCore.QPoint(x2,y2)))
+
+
+                if self.parent().checkBox_3.isChecked():
+                    if box in active_boxes:
+                        qp.drawEllipse(QtCore.QPoint(self.zoom_factor*(self.gox + x+4),self.zoom_factor*(self.goy + y+4)),6, 6)
+                    else:
+                        qp.drawEllipse(QtCore.QPoint(self.zoom_factor*(self.gox + x+4),self.zoom_factor*(self.goy + y+4)),2, 2)
+
 
-                #qp.drawRect(QtCore.QRect(QtCore.QPoint(x,y),QtCore.QPoint(x2,y2)))
-                qp.drawEllipse(QtCore.QPoint(self.zoom_factor*(self.gox + x+3),self.zoom_factor*(self.goy + y+3)),6, 6)
 
                 """
                 if self.darkMode:
@@ -927,6 +1116,9 @@ class MyDrawWidget(QtWidgets.QWidget):
 
         self.parent().update_graphics_data(self.boxes,self.connectors)
 
+        time_delay = time.time() - self.last_click
+        self.last_click = time.time()
+
 
         if event.buttons() & QtCore.Qt.RightButton:
             self.mode = "drag"
@@ -964,17 +1156,24 @@ class MyDrawWidget(QtWidgets.QWidget):
                     self.offy = (box.y - y)
 
                 elif self.mode == "select":
-                    if self.highlighted == box:
+                    if time_delay < 0.2: # double click on agent to edit
+                        if self.highlighted is None:
+                            self.highlighted = box
+                        self.edit_agent()
                         self.highlighted = None
-                        print("unhighlight", box)
+                        return
                     else:
-                        print("highlight",box)
-                        self.highlighted = box
+                        if self.highlighted == box:
+                            self.highlighted = None
+                            print("unhighlight", box)
+                            self.update()
+                        else:
+                            print("highlight",box)
+                            self.highlighted = box
 
+                    # self.mode = "drag"
 
-                    self.mode = "drag"
-
-                    if box.ishelper:
+                    if box.ishelper and self.highlighted == box:
                         self.parent().data_select_where(box.name)
 
                     self.update()
@@ -1105,8 +1304,8 @@ class MyDrawWidget(QtWidgets.QWidget):
 
             if self.selected is not None:
 
-                self.selected.x = np.clip(self.mousex + self.offx,1-self.gox,self.maxx-self.gox)
-                self.selected.y = np.clip(self.mousey + self.offy, 1-self.goy, self.maxy-self.goy)
+                self.selected.x = self.mousex + self.offx # np.clip(self.mousex + self.offx,1-self.gox,2*self.maxx-self.gox)
+                self.selected.y = self.mousey + self.offy # np.clip(self.mousey + self.offy, 1-self.goy,2*self.maxy-self.goy)
                 self.selected.x0 = self.selected.x
                 self.selected.y0 = self.selected.y
 
